import io
import json
import os
import random
import shutil
import sys
import tempfile
import traceback
import unittest
from contextlib import contextmanager
from datetime import datetime
from functools import wraps
from unittest.mock import patch

import dbt.flags as flags
import dbt.main as dbt
import pytest
import yaml
from dbt.adapters.factory import get_adapter, reset_adapters, register_adapter
from dbt.clients.jinja import template_cache
from dbt.config import RuntimeConfig
from dbt.context import providers
from dbt.deprecations import reset_deprecations
from dbt.logger import GLOBAL_LOGGER as logger, log_manager

from dbt.adapters.oracle import OracleAdapter

INITIAL_ROOT = os.getcwd()


def normalize(path):
    """On windows, neither is enough on its own:

    >>> normcase('C:\\documents/ALL CAPS/subdir\\..')
    'c:\\documents\\all caps\\subdir\\..'
    >>> normpath('C:\\documents/ALL CAPS/subdir\\..')
    'C:\\documents\\ALL CAPS'
    >>> normpath(normcase('C:\\documents/ALL CAPS/subdir\\..'))
    'c:\\documents\\all caps'
    """
    return os.path.normcase(os.path.normpath(path))


class Normalized:
    def __init__(self, value):
        self.value = value

    def __repr__(self):
        return f'Normalized({self.value!r})'

    def __str__(self):
        return f'Normalized({self.value!s})'

    def __eq__(self, other):
        return normalize(self.value) == normalize(other)


class FakeArgs:
    def __init__(self):
        self.threads = 1
        self.data = False
        self.defer = False
        self.schema = True
        self.full_refresh = False
        self.models = None
        self.exclude = None
        self.single_threaded = False
        self.selector_name = None
        self.state = None
        self.defer = None


class TestArgs:
    def __init__(self, kwargs):
        self.which = 'run'
        self.single_threaded = False
        self.profiles_dir = None
        self.project_dir = None
        self.__dict__.update(kwargs)


def _profile_from_test_name(test_name):
    adapter_names = ('postgres', 'snowflake', 'redshift', 'bigquery', 'presto')
    adapters_in_name = sum(x in test_name for x in adapter_names)
    if adapters_in_name != 1:
        raise ValueError(
            'test names must have exactly 1 profile choice embedded, {} has {}'
                .format(test_name, adapters_in_name)
        )

    for adapter_name in adapter_names:
        if adapter_name in test_name:
            return adapter_name

    raise ValueError(
        'could not find adapter name in test name {}'.format(test_name)
    )


def _pytest_test_name():
    return os.environ['PYTEST_CURRENT_TEST'].split()[0]


def _pytest_get_test_root():
    test_path = _pytest_test_name().split('::')[0]
    relative_to = INITIAL_ROOT
    head = os.path.relpath(test_path, relative_to)

    path_parts = []
    while head:
        head, tail = os.path.split(head)
        path_parts.append(tail)
    path_parts.reverse()
    # dbt tests are all of the form 'test/integration/XXX_suite_name'
    target = os.path.join(*path_parts[:3])
    return os.path.join(relative_to, target)


def _really_makedirs(path):
    while not os.path.exists(path):
        try:
            os.makedirs(path)
        except EnvironmentError:
            raise


class DBTIntegrationTest(unittest.TestCase):
    CREATE_SCHEMA_STATEMENT = 'CREATE SCHEMA {}'
    DROP_SCHEMA_STATEMENT = 'DROP SCHEMA IF EXISTS {} CASCADE'

    # ORACLE CUSTOM STATEMENTS
    CREATE_SCHEMA_STATEMENT_ORACLE = 'CREATE USER {0} IDENTIFIED BY {0}'
    CREATE_SCHEMA_STATEMENT_ORACLE_GRANT = 'GRANT ALL PRIVILEGES TO {0}'

    DROP_SCHEMA_STATEMENT = """
        BEGIN
            EXECUTE IMMEDIATE 'DROP USER {} CASCADE';
        EXCEPTION WHEN OTHERS THEN NULL;
        END;
    """

    _randint = random.randint(0, 9999)
    _runtime_timedelta = (datetime.utcnow() - datetime(1970, 1, 1, 0, 0, 0))
    _runtime = (
            (int(_runtime_timedelta.total_seconds() * 1e6)) +
            _runtime_timedelta.microseconds
    )

    # prefix = f'test{_runtime}{_randint:04}'
    # ORACLE FIX
    prefix = f'test{_runtime}{_randint:04}'[:10]
    setup_alternate_db = False

    def oracle_safe_drop_table(self, schema, table):
        self.run_sql(f"BEGIN EXECUTE IMMEDIATE 'DROP TABLE {schema}.{table}'; EXCEPTION WHEN OTHERS THEN NULL; END;")

    def oracle_safe_drop_user(self, user):
        self.run_sql(f"BEGIN EXECUTE IMMEDIATE 'DROP USER {user} CASCADE'; EXCEPTION WHEN OTHERS THEN NULL; END;")

    @property
    def database_host(self):
        if os.name == 'nt':
            return 'localhost'
        return 'database'

    def postgres_profile(self):
        return self.oracle_profile()  # override postgres with oracle profile
        # return {
        #     'config': {
        #         'send_anonymous_usage_stats': False
        #     },
        #     'test': {
        #         'outputs': {
        #             'default2': {
        #                 'type': 'postgres',
        #                 'threads': 4,
        #                 'host': self.database_host,
        #                 'port': 5432,
        #                 'user': 'root',
        #                 'pass': 'password',
        #                 'dbname': 'dbt',
        #                 'schema': self.unique_schema()
        #             },
        #             'noaccess': {
        #                 'type': 'postgres',
        #                 'threads': 4,
        #                 'host': self.database_host,
        #                 'port': 5432,
        #                 'user': 'noaccess',
        #                 'pass': 'password',
        #                 'dbname': 'dbt',
        #                 'schema': self.unique_schema()
        #             }
        #         },
        #         'target': 'default2'
        #     }
        # }

    def oracle_profile(self):
        return {
            'config': {
                'send_anonymous_usage_stats': False
            },
            'test': {
                'outputs': {
                    'default2': {
                        'type': 'oracle',
                        'threads': 4,
                        'host': 'localhost',
                        'port': 1521,
                        'service': 'XEPDB1',
                        'username': 'SYS',
                        'password': 'root',
                        'as_sysdba': True,
                        'nls_date_format': 'YYYY-MM-DD HH24:MI:SS',
                        'schema': self.unique_schema()
                    },
                    'noaccess': {
                        'type': 'oracle',
                        'threads': 4,
                        'host': 'localhost',
                        'port': 1521,
                        'service': 'XEPDB1',
                        'username': 'noaccess',
                        'password': 'noaccess',
                        'schema': self.unique_schema()
                    }
                },
                'target': 'default2'
            }
        }

    def redshift_profile(self):
        return {
            'config': {
                'send_anonymous_usage_stats': False
            },
            'test': {
                'outputs': {
                    'default2': {
                        'type': 'redshift',
                        'threads': 1,
                        'host': os.getenv('REDSHIFT_TEST_HOST'),
                        'port': int(os.getenv('REDSHIFT_TEST_PORT')),
                        'user': os.getenv('REDSHIFT_TEST_USER'),
                        'pass': os.getenv('REDSHIFT_TEST_PASS'),
                        'dbname': os.getenv('REDSHIFT_TEST_DBNAME'),
                        'schema': self.unique_schema()
                    }
                },
                'target': 'default2'
            }
        }

    def snowflake_profile(self):
        return {
            'config': {
                'send_anonymous_usage_stats': False
            },
            'test': {
                'outputs': {
                    'default2': {
                        'type': 'snowflake',
                        'threads': 4,
                        'account': os.getenv('SNOWFLAKE_TEST_ACCOUNT'),
                        'user': os.getenv('SNOWFLAKE_TEST_USER'),
                        'password': os.getenv('SNOWFLAKE_TEST_PASSWORD'),
                        'database': os.getenv('SNOWFLAKE_TEST_DATABASE'),
                        'schema': self.unique_schema(),
                        'warehouse': os.getenv('SNOWFLAKE_TEST_WAREHOUSE'),
                    },
                    'noaccess': {
                        'type': 'snowflake',
                        'threads': 4,
                        'account': os.getenv('SNOWFLAKE_TEST_ACCOUNT'),
                        'user': 'noaccess',
                        'password': 'password',
                        'database': os.getenv('SNOWFLAKE_TEST_DATABASE'),
                        'schema': self.unique_schema(),
                        'warehouse': os.getenv('SNOWFLAKE_TEST_WAREHOUSE'),
                    },
                    'oauth': {
                        'type': 'snowflake',
                        'threads': 4,
                        'account': os.getenv('SNOWFLAKE_TEST_ACCOUNT'),
                        'user': os.getenv('SNOWFLAKE_TEST_USER'),
                        'oauth_client_id': os.getenv('SNOWFLAKE_TEST_OAUTH_CLIENT_ID'),
                        'oauth_client_secret': os.getenv('SNOWFLAKE_TEST_OAUTH_CLIENT_SECRET'),
                        'token': os.getenv('SNOWFLAKE_TEST_OAUTH_REFRESH_TOKEN'),
                        'database': os.getenv('SNOWFLAKE_TEST_DATABASE'),
                        'schema': self.unique_schema(),
                        'warehouse': os.getenv('SNOWFLAKE_TEST_WAREHOUSE'),
                        'authenticator': 'oauth',
                    },
                },
                'target': 'default2'
            }
        }

    def bigquery_profile(self):
        credentials_json_str = os.getenv('BIGQUERY_SERVICE_ACCOUNT_JSON').replace("'", '')
        credentials = json.loads(credentials_json_str)
        project_id = credentials.get('project_id')

        return {
            'config': {
                'send_anonymous_usage_stats': False
            },
            'test': {
                'outputs': {
                    'default2': {
                        'type': 'bigquery',
                        'method': 'service-account-json',
                        'threads': 1,
                        'project': project_id,
                        'keyfile_json': credentials,
                        'schema': self.unique_schema(),
                    },
                },
                'target': 'default2'
            }
        }

    def presto_profile(self):
        return {
            'config': {
                'send_anonymous_usage_stats': False
            },
            'test': {
                'outputs': {
                    'default2': {
                        'type': 'presto',
                        'method': 'none',
                        'threads': 1,
                        'schema': self.unique_schema(),
                        'database': 'hive',
                        'host': 'presto',
                        'port': 8080,
                    },
                },
                'target': 'default2'
            }
        }

    @property
    def packages_config(self):
        return None

    @property
    def selectors_config(self):
        return None

    def unique_schema(self):
        schema = self.schema[:19]

        to_return = "{}_{}".format(self.prefix, schema)

        if self.adapter_type == 'snowflake':
            return to_return.upper()

        return to_return.lower()

    @property
    def default_database(self):
        database = self.config.credentials.database
        if self.adapter_type == 'snowflake':
            return database.upper()
        return database

    @property
    def alternative_database(self):
        if self.adapter_type == 'bigquery':
            return os.environ['BIGQUERY_TEST_ALT_DATABASE']
        elif self.adapter_type == 'snowflake':
            return os.environ['SNOWFLAKE_TEST_ALT_DATABASE']
        return None

    def get_profile(self, adapter_type):
        if adapter_type == 'postgres':
            return self.postgres_profile()
        elif adapter_type == 'snowflake':
            return self.snowflake_profile()
        elif adapter_type == 'bigquery':
            return self.bigquery_profile()
        elif adapter_type == 'redshift':
            return self.redshift_profile()
        elif adapter_type == 'presto':
            return self.presto_profile()
        else:
            raise ValueError('invalid adapter type {}'.format(adapter_type))

    def _pick_profile(self):
        test_name = self.id().split('.')[-1]
        return _profile_from_test_name(test_name)

    def _symlink_test_folders(self):
        for entry in os.listdir(self.test_original_source_path):
            src = os.path.join(self.test_original_source_path, entry)
            tst = os.path.join(self.test_root_dir, entry)
            if os.path.isdir(src) or src.endswith('.sql'):
                # symlink all sql files and all directories.
                os.symlink(src, tst)
        os.symlink(self._logs_dir, os.path.join(self.test_root_dir, 'logs'))

    @property
    def test_root_realpath(self):
        if sys.platform == 'darwin':
            return os.path.realpath(self.test_root_dir)
        else:
            return self.test_root_dir

    def _generate_test_root_dir(self):
        return normalize(tempfile.mkdtemp(prefix='dbt-int-test-'))

    def setUp(self):
        self.dbt_core_install_root = os.path.dirname(dbt.__file__)
        log_manager.reset_handlers()
        self.initial_dir = INITIAL_ROOT
        os.chdir(self.initial_dir)
        # before we go anywhere, collect the initial path info
        self._logs_dir = os.path.join(self.initial_dir, 'logs', self.prefix)
        _really_makedirs(self._logs_dir)
        self.test_original_source_path = _pytest_get_test_root()
        self.test_root_dir = self._generate_test_root_dir()

        os.chdir(self.test_root_dir)
        try:
            self._symlink_test_folders()
        except Exception as exc:
            msg = '\n\t'.join((
                'Failed to symlink test folders!',
                'initial_dir={0.initial_dir}',
                'test_original_source_path={0.test_original_source_path}',
                'test_root_dir={0.test_root_dir}'
            )).format(self)
            logger.exception(msg)

            # if logging isn't set up, I still really want this message.
            print(msg)
            traceback.print_exc()

            raise

        self._created_schemas = set()
        reset_deprecations()
        flags.reset()
        template_cache.clear()

        self.use_profile(self._pick_profile())
        self.use_default_project()
        self.set_packages()
        self.set_selectors()
        self.load_config()

    def use_default_project(self, overrides=None):
        # create a dbt_project.yml
        base_project_config = {
            'name': 'test',
            'version': '1.0',
            'config-version': 2,
            'test-paths': [],
            'source-paths': [self.models],
            'profile': 'test',
        }

        project_config = {}
        project_config.update(base_project_config)
        project_config.update(self.project_config)
        project_config.update(overrides or {})

        with open("dbt_project.yml", 'w') as f:
            yaml.safe_dump(project_config, f, default_flow_style=True)

    def use_profile(self, adapter_type):
        self.adapter_type = adapter_type

        profile_config = {}
        default_profile_config = self.get_profile(adapter_type)

        profile_config.update(default_profile_config)
        profile_config.update(self.profile_config)

        if not os.path.exists(self.test_root_dir):
            os.makedirs(self.test_root_dir)

        profiles_path = os.path.join(self.test_root_dir, 'profiles.yml')
        with open(profiles_path, 'w') as f:
            yaml.safe_dump(profile_config, f, default_flow_style=True)
        self._profile_config = profile_config

    def set_packages(self):
        if self.packages_config is not None:
            with open('packages.yml', 'w') as f:
                yaml.safe_dump(self.packages_config, f, default_flow_style=True)

    def set_selectors(self):
        if self.selectors_config is not None:
            with open('selectors.yml', 'w') as f:
                yaml.safe_dump(self.selectors_config, f, default_flow_style=True)

    def load_config(self):
        # we've written our profile and project. Now we want to instantiate a
        # fresh adapter for the tests.
        # it's important to use a different connection handle here so
        # we don't look into an incomplete transaction
        kwargs = {
            'profile': None,
            'profiles_dir': self.test_root_dir,
            'target': None,
        }

        config = RuntimeConfig.from_args(TestArgs(kwargs))

        register_adapter(config)
        adapter = get_adapter(config)
        adapter.cleanup_connections()
        self.adapter_type = adapter.type()
        self.adapter = adapter
        self.config = config

        self._drop_schemas()
        self._create_schemas()

    def quote_as_configured(self, value, quote_key):
        # return self.adapter.quote_as_configured(value, quote_key)
        return value  # never quote on oracle

    def tearDown(self):
        # get any current run adapter and clean up its connections before we
        # reset them. It'll probably be different from ours because
        # handle_and_check() calls reset_adapters().
        register_adapter(self.config)
        adapter = get_adapter(self.config)
        if adapter is not self.adapter:
            adapter.cleanup_connections()
        if not hasattr(self, 'adapter'):
            self.adapter = adapter

        self._drop_schemas()

        self.adapter.cleanup_connections()
        reset_adapters()
        os.chdir(INITIAL_ROOT)
        try:
            shutil.rmtree(self.test_root_dir)
        except EnvironmentError:
            logger.exception('Could not clean up after test - {} not removable'
                             .format(self.test_root_dir))

    def _get_schema_fqn(self, database, schema):
        schema_fqn = self.quote_as_configured(schema, 'schema')
        if self.adapter_type == 'snowflake':
            database = self.quote_as_configured(database, 'database')
            schema_fqn = '{}.{}'.format(database, schema_fqn)
        return schema_fqn

    def _create_schema_named(self, database, schema):
        if self.adapter_type == 'bigquery':
            relation = self.adapter.Relation.create(database=database, schema=schema)
            self.adapter.create_schema(relation)
        else:
            schema_fqn = self._get_schema_fqn(database, schema)
            # self.run_sql(self.CREATE_SCHEMA_STATEMENT.format(schema_fqn))
            self.run_sql(self.CREATE_SCHEMA_STATEMENT_ORACLE.format(schema_fqn))
            self.run_sql(self.CREATE_SCHEMA_STATEMENT_ORACLE_GRANT.format(schema_fqn))
            self._created_schemas.add(schema_fqn)

    def _drop_schema_named(self, database, schema):
        if self.adapter_type == 'bigquery' or self.adapter_type == 'presto':
            relation = self.adapter.Relation.create(database=database, schema=schema)
            self.adapter.drop_schema(relation)
        else:
            schema_fqn = self._get_schema_fqn(database, schema)
            self.run_sql(self.DROP_SCHEMA_STATEMENT.format(schema_fqn))

    def _create_schemas(self):
        schema = self.unique_schema()
        with self.adapter.connection_named('__test'):
            self._create_schema_named(self.default_database, schema)
            if self.setup_alternate_db and self.adapter_type == 'snowflake':
                self._create_schema_named(self.alternative_database, schema)

    def _drop_schemas_adapter(self):
        schema = self.unique_schema()
        if self.adapter_type == 'bigquery' or self.adapter_type == 'presto':
            self._drop_schema_named(self.default_database, schema)
            if self.setup_alternate_db and self.alternative_database:
                self._drop_schema_named(self.alternative_database, schema)

    def _drop_schemas_sql(self):
        schema = self.unique_schema()
        # we always want to drop these if necessary, we'll clear it soon.
        self._created_schemas.add(
            self._get_schema_fqn(self.default_database, schema)
        )
        # on postgres/redshift, this will make you sad
        drop_alternative = (
                self.setup_alternate_db and
                self.adapter_type not in {'postgres', 'redshift'} and
                self.alternative_database
        )
        if drop_alternative:
            self._created_schemas.add(
                self._get_schema_fqn(self.alternative_database, schema)
            )

        for schema_fqn in self._created_schemas:
            self.run_sql(self.DROP_SCHEMA_STATEMENT.format(schema_fqn))

        self._created_schemas.clear()

    def _drop_schemas(self):
        with self.adapter.connection_named('__test'):
            if self.adapter_type == 'bigquery' or self.adapter_type == 'presto':
                self._drop_schemas_adapter()
            else:
                self._drop_schemas_sql()

    @property
    def project_config(self):
        return {
            'config-version': 2,
        }

    @property
    def profile_config(self):
        return {}

    def run_dbt(self, args=None, expect_pass=True, strict=True, parser=True, profiles_dir=True):
        res, success = self.run_dbt_and_check(args=args, strict=strict, parser=parser, profiles_dir=profiles_dir)

        self.assertEqual(
            success, expect_pass,
            "dbt exit state did not match expected")

        return res

    def run_dbt_and_capture(self, *args, **kwargs):
        try:
            initial_stdout = log_manager.stdout
            initial_stderr = log_manager.stderr
            stringbuf = io.StringIO()
            log_manager.set_output_stream(stringbuf)

            res = self.run_dbt(*args, **kwargs)
            stdout = stringbuf.getvalue()

        finally:
            log_manager.set_output_stream(initial_stdout, initial_stderr)

        return res, stdout

    def run_dbt_and_check(self, args=None, strict=True, parser=False, profiles_dir=True):
        log_manager.reset_handlers()
        if args is None:
            args = ["run"]

        final_args = []

        final_args.append('--debug')  # easier to debug tests

        if strict:
            final_args.append('--strict')
        if parser:
            final_args.append('--test-new-parser')
        if os.getenv('DBT_TEST_SINGLE_THREADED') in ('y', 'Y', '1'):
            final_args.append('--single-threaded')

        final_args.extend(args)

        if profiles_dir:
            final_args.extend(['--profiles-dir', self.test_root_dir])
        final_args.append('--log-cache-events')

        logger.info("Invoking dbt with {}".format(final_args))
        return dbt.handle_and_check(final_args)

    def run_sql_file(self, path, kwargs=None):
        with open(path, 'r') as f:
            statements = f.read().split(";")
            for statement in statements:
                self.run_sql(statement, kwargs=kwargs)

    # horrible hack to support snowflake for right now
    def transform_sql(self, query, kwargs=None):
        to_return = query

        if self.adapter_type == 'snowflake':
            to_return = to_return.replace("BIGSERIAL", "BIGINT AUTOINCREMENT")

        base_kwargs = {
            'schema': self.unique_schema(),
            'database': self.adapter.quote(self.default_database),
        }
        if kwargs is None:
            kwargs = {}
        base_kwargs.update(kwargs)

        to_return = to_return.format(**base_kwargs)

        return to_return

    def run_sql_bigquery(self, sql, fetch):
        """Run an SQL query on a bigquery adapter. No cursors, transactions,
        etc. to worry about"""

        do_fetch = fetch != 'None'
        _, res = self.adapter.execute(sql, fetch=do_fetch)

        # convert dataframe to matrix-ish repr
        if fetch == 'one':
            return res[0]
        else:
            return list(res)

    def run_sql_presto(self, sql, fetch, conn):
        cursor = conn.handle.cursor()
        try:
            cursor.execute(sql)
            if fetch == 'one':
                return cursor.fetchall()[0]
            elif fetch == 'all':
                return cursor.fetchall()
            else:
                # we have to fetch.
                cursor.fetchall()
        except Exception as e:
            conn.handle.rollback()
            conn.transaction_open = False
            print(sql)
            print(e)
            raise
        else:
            conn.handle.commit()
            conn.transaction_open = False

    def run_sql_common(self, sql_raw, fetch, conn):
        # ORACLE FIX
        sql_list = OracleAdapter._oracle_query_parse(sql_raw)
        final_result = None
        for sql in sql_list:
            with conn.handle.cursor() as cursor:
                try:
                    cursor.execute(sql)
                    conn.handle.commit()
                    if fetch == 'one':
                        final_result = cursor.fetchone()
                    elif fetch == 'all':
                        final_result = cursor.fetchall()
                    else:
                        final_result = None
                except BaseException as e:
                    if conn.handle and not getattr(conn.handle, 'closed', True):
                        conn.handle.rollback()
                    print(sql)
                    print(e)
                    raise
                finally:
                    conn.transaction_open = False
        return final_result

    def run_sql(self, query, fetch='None', kwargs=None, connection_name=None):
        if connection_name is None:
            connection_name = '__test'

        if query.strip() == "":
            return

        sql = self.transform_sql(query, kwargs=kwargs)

        with self.get_connection(connection_name) as conn:
            logger.debug('test connection "{}" executing: {}'.format(conn.name, sql))
            if self.adapter_type == 'bigquery':
                return self.run_sql_bigquery(sql, fetch)
            elif self.adapter_type == 'presto':
                return self.run_sql_presto(sql, fetch, conn)
            else:
                return self.run_sql_common(sql, fetch, conn)

    def _ilike(self, target, value):
        # presto has this regex substitution monstrosity instead of 'ilike'
        if self.adapter_type == 'presto':
            return r"regexp_like({}, '(?i)\A{}\Z')".format(target, value)
        else:
            #  return "{} ilike '{}'".format(target, value)
            # ORACLE STATEMENT
            return "UPPER({}) LIKE UPPER('%{}%')".format(target, value)

    def get_many_table_columns_snowflake(self, tables, schema, database=None):
        tables = set(tables)
        if database is None:
            database = self.default_database
        sql = 'show columns in schema {database}.{schema}'.format(
            database=self.quote_as_configured(database, 'database'),
            schema=self.quote_as_configured(schema, 'schema')
        )
        # assumption: this will be much  faster than doing one query/table
        # because in tests, we'll want most of our tables most of the time.
        columns = self.run_sql(sql, fetch='all')
        results = []
        for column in columns:
            table_name, _, column_name, json_data_type = column[:4]
            character_maximum_length = None
            if table_name in tables:
                typeinfo = json.loads(json_data_type)
                data_type = typeinfo['type']
                if data_type == 'TEXT':
                    character_maximum_length = max(typeinfo['length'], 16777216)
                results.append((table_name, column_name, data_type, character_maximum_length))
        return results

    def get_many_table_columns_information_schema(self, tables, schema, database=None):
        if self.adapter_type == 'presto':
            columns = 'table_name, column_name, data_type'
        else:
            columns = 'table_name, column_name, data_type, character_maximum_length'

        sql = """
                select {columns}
                from {db_string}information_schema.columns
                where {schema_filter}
                  and ({table_filter})
                order by column_name asc"""

        db_string = ''
        if database:
            db_string = self.quote_as_configured(database, 'database') + '.'

        table_filters_s = " OR ".join(
            self._ilike('table_name', table.replace('"', ''))
            for table in tables
        )
        schema_filter = self._ilike('table_schema', schema)

        # ORACLE STATEMENT
        columns = 'table_name, column_name, data_type, char_length'
        sql = """
            SELECT
                {columns}
            FROM
                ALL_TAB_COLS
            WHERE
                {schema_filter}
                AND ({table_filter})
            ORDER BY
                column_name ASc
        """
        schema_filter = self._ilike('owner', schema)

        sql = sql.format(
            columns=columns,
            schema_filter=schema_filter,
            table_filter=table_filters_s,
            db_string=db_string)

        columns = self.run_sql(sql, fetch='all')
        return list(map(self.filter_many_columns, columns))

    def get_many_table_columns_bigquery(self, tables, schema, database=None):
        result = []
        for table in tables:
            relation = self._make_relation(table, schema, database)
            columns = self.adapter.get_columns_in_relation(relation)
            for col in columns:
                result.append((table, col.column, col.dtype, col.char_size))
        return result

    def get_many_table_columns(self, tables, schema, database=None):
        if self.adapter_type == 'snowflake':
            result = self.get_many_table_columns_snowflake(tables, schema, database)
        elif self.adapter_type == 'bigquery':
            result = self.get_many_table_columns_bigquery(tables, schema, database)
        else:
            result = self.get_many_table_columns_information_schema(tables, schema, database)
        result.sort(key=lambda x: '{}.{}'.format(x[0], x[1]))
        return result

    def filter_many_columns(self, column):
        if len(column) == 3:
            table_name, column_name, data_type = column
            char_size = None
        else:
            table_name, column_name, data_type, char_size = column
        # in snowflake, all varchar widths are created equal
        if self.adapter_type == 'snowflake':
            if char_size and char_size < 16777216:
                char_size = 16777216
        return (table_name, column_name, data_type, char_size)

    @contextmanager
    def get_connection(self, name=None):
        """Create a test connection context where all executed macros, etc will
        get self.adapter as the adapter.

        This allows tests to run normal adapter macros as if reset_adapters()
        were not called by handle_and_check (for asserts, etc)
        """
        if name is None:
            name = '__test'
        with patch.object(providers, 'get_adapter', return_value=self.adapter):
            with self.adapter.connection_named(name):
                conn = self.adapter.connections.get_thread_connection()
                yield conn

    def get_relation_columns(self, relation):
        with self.get_connection():
            columns = self.adapter.get_columns_in_relation(relation)

        return sorted(((c.name, c.dtype, c.char_size) for c in columns),
                      key=lambda x: x[0])

    def get_table_columns(self, table, schema=None, database=None):
        schema = self.unique_schema() if schema is None else schema
        database = self.default_database if database is None else database
        relation = self.adapter.Relation.create(
            database=database,
            schema=schema,
            identifier=table,
            type='table',
            quote_policy=self.config.quoting
        )
        return self.get_relation_columns(relation)

    def get_table_columns_as_dict(self, tables, schema=None):
        col_matrix = self.get_many_table_columns(tables, schema)
        res = {}
        for row in col_matrix:
            table_name = row[0]
            col_def = row[1:]
            if table_name not in res:
                res[table_name] = []
            res[table_name].append(col_def)
        return res

    def get_models_in_schema_snowflake(self, schema):
        sql = 'show objects in schema {}.{}'.format(
            self.quote_as_configured(self.default_database, 'database'),
            self.quote_as_configured(schema, 'schema')
        )
        results = {}
        for row in self.run_sql(sql, fetch='all'):
            # I sure hope these never change!
            name = row[1]
            kind = row[4]

            if kind == 'TABLE':
                kind = 'table'
            elif kind == 'VIEW':
                kind = 'view'

            results[name] = kind
        return results

    def get_models_in_schema(self, schema=None):
        schema = self.unique_schema() if schema is None else schema
        if self.adapter_type == 'snowflake':
            return self.get_models_in_schema_snowflake(schema)

        sql = """
                select table_name,
                        case when table_type = 'BASE TABLE' then 'table'
                             when table_type = 'VIEW' then 'view'
                             else table_type
                        end as materialization
                from information_schema.tables
                where {}
                order by table_name
                """
        sql = sql.format(self._ilike('table_schema', schema))

        # ORACLE STATEMENT
        sql = """
                SELECT
                    table_name,
                    'table' AS materialization
                FROM
                    ALL_TABLES
                WHERE
                    {0}
                UNION ALL
                SELECT
                    view_name,
                    'view' AS materialization
                FROM
                    ALL_VIEWS
                WHERE
                    {0}
        """
        sql = sql.format(self._ilike('owner', schema))

        result = self.run_sql(sql, fetch='all')

        return {str(model_name).lower(): materialization for (model_name, materialization) in result}

    def _assertTablesEqualSql(self, relation_a, relation_b, columns=None):
        if columns is None:
            columns = self.get_relation_columns(relation_a)
        column_names = [c[0] for c in columns]

        sql = self.adapter.get_rows_different_sql(
            relation_a, relation_b, column_names
        )

        return sql

    def assertTablesEqual(self, table_a, table_b,
                          table_a_schema=None, table_b_schema=None,
                          table_a_db=None, table_b_db=None):
        if table_a_schema is None:
            table_a_schema = self.unique_schema()

        if table_b_schema is None:
            table_b_schema = self.unique_schema()

        if table_a_db is None:
            table_a_db = self.default_database

        if table_b_db is None:
            table_b_db = self.default_database

        relation_a = self._make_relation(table_a, table_a_schema, table_a_db)
        relation_b = self._make_relation(table_b, table_b_schema, table_b_db)

        self._assertTableColumnsEqual(relation_a, relation_b)

        sql = self._assertTablesEqualSql(relation_a, relation_b)
        result = self.run_sql(sql, fetch='one')

        self.assertEqual(
            result[0],
            0,
            'row_count_difference nonzero: ' + sql
        )
        self.assertEqual(
            result[1],
            0,
            'num_mismatched nonzero: ' + sql
        )

    def _make_relation(self, identifier, schema=None, database=None):
        if schema is None:
            schema = self.unique_schema()
        if database is None:
            database = self.default_database
        return self.adapter.Relation.create(
            database=database,
            schema=schema,
            identifier=identifier,
            quote_policy=self.config.quoting
        )

    def get_many_relation_columns(self, relations):
        """Returns a dict of (datbase, schema) -> (dict of (table_name -> list of columns))
        """
        schema_fqns = {}
        for rel in relations:
            this_schema = schema_fqns.setdefault((rel.database, rel.schema), [])
            this_schema.append(rel.identifier)

        column_specs = {}
        for key, tables in schema_fqns.items():
            database, schema = key
            columns = self.get_many_table_columns(tables, schema, database=database)
            table_columns = {}
            for col in columns:
                table_columns.setdefault(col[0], []).append(col[1:])
            for rel_name, columns in table_columns.items():
                key = (database, schema, rel_name)
                column_specs[key] = columns

        return column_specs

    def assertManyRelationsEqual(self, relations, default_schema=None, default_database=None):
        if default_schema is None:
            default_schema = self.unique_schema()
        if default_database is None:
            default_database = self.default_database

        specs = []
        for relation in relations:
            if not isinstance(relation, (tuple, list)):
                relation = [relation]

            assert len(relation) <= 3

            if len(relation) == 3:
                relation = self._make_relation(*relation)
            elif len(relation) == 2:
                relation = self._make_relation(relation[0], relation[1], default_database)
            elif len(relation) == 1:
                relation = self._make_relation(relation[0], default_schema, default_database)
            else:
                raise ValueError('relation must be a sequence of 1, 2, or 3 values')

            specs.append(relation)

        with self.get_connection():
            column_specs = self.get_many_relation_columns(specs)

        # make sure everyone has equal column definitions
        first_columns = None
        for relation in specs:
            key = (relation.database, relation.schema, relation.identifier)
            # get a good error here instead of a hard-to-diagnose KeyError
            self.assertIn(key, column_specs, f'No columns found for {key}')
            columns = column_specs[key]
            if first_columns is None:
                first_columns = columns
            else:
                self.assertEqual(
                    first_columns, columns,
                    '{} did not match {}'.format(str(specs[0]), str(relation))
                )

        # make sure everyone has the same data. if we got here, everyone had
        # the same column specs!
        first_relation = None
        for relation in specs:
            if first_relation is None:
                first_relation = relation
            else:
                sql = self._assertTablesEqualSql(first_relation, relation,
                                                 columns=first_columns)
                result = self.run_sql(sql, fetch='one')

                self.assertEqual(
                    result[0],
                    0,
                    'row_count_difference nonzero: ' + sql
                )
                self.assertEqual(
                    result[1],
                    0,
                    'num_mismatched nonzero: ' + sql
                )

    def assertManyTablesEqual(self, *args):
        schema = self.unique_schema()

        all_tables = []
        for table_equivalencies in args:
            all_tables += list(table_equivalencies)

        all_cols = self.get_table_columns_as_dict(all_tables, schema)

        for table_equivalencies in args:
            first_table = table_equivalencies[0]

            # ORACLE FIX
            first_table = first_table.upper()

            first_relation = self._make_relation(first_table)

            # assert that all tables have the same columns
            base_result = all_cols[first_table]
            self.assertTrue(len(base_result) > 0)

            for other_table in table_equivalencies[1:]:
                # ORACLE FIX
                other_table = other_table.upper()

                other_result = all_cols[other_table]
                self.assertTrue(len(other_result) > 0)
                self.assertEqual(base_result, other_result)

                other_relation = self._make_relation(other_table)
                sql = self._assertTablesEqualSql(first_relation,
                                                 other_relation,
                                                 columns=base_result)
                result = self.run_sql(sql, fetch='one')

                self.assertEqual(
                    result[0],
                    0,
                    'row_count_difference nonzero: ' + sql
                )
                self.assertEqual(
                    result[1],
                    0,
                    'num_mismatched nonzero: ' + sql
                )

    def _assertTableRowCountsEqual(self, relation_a, relation_b):
        cmp_query = """
            with table_a as (

                select count(*) as num_rows from {}

            ), table_b as (

                select count(*) as num_rows from {}

            )

            select table_a.num_rows - table_b.num_rows as difference
            from table_a, table_b

        """.format(str(relation_a), str(relation_b))

        res = self.run_sql(cmp_query, fetch='one')

        self.assertEqual(int(res[0]), 0,
                         "Row count of table {} doesn't match row count of table {}. ({} rows different)".format(
                             relation_a.identifier,
                             relation_b.identifier,
                             res[0]
                         )
                         )

    def assertTableDoesNotExist(self, table, schema=None, database=None):
        columns = self.get_table_columns(table, schema, database)

        self.assertEqual(
            len(columns),
            0
        )

    def assertTableDoesExist(self, table, schema=None, database=None):
        columns = self.get_table_columns(table, schema, database)

        self.assertGreater(
            len(columns),
            0
        )

    def _assertTableColumnsEqual(self, relation_a, relation_b):
        table_a_result = self.get_relation_columns(relation_a)
        table_b_result = self.get_relation_columns(relation_b)

        self.assertEqual(len(table_a_result), len(table_b_result))
        for a_column, b_column in zip(table_a_result, table_b_result):
            a_name, a_type, a_size = a_column
            b_name, b_type, b_size = b_column
            self.assertEqual(a_name, b_name,
                             '{} vs {}: column "{}" != "{}"'.format(
                                 relation_a, relation_b, a_name, b_name
                             ))

            self.assertEqual(a_type, b_type,
                             '{} vs {}: column "{}" has type "{}" != "{}"'.format(
                                 relation_a, relation_b, a_name, a_type, b_type
                             ))

            if self.adapter_type == 'presto' and None in (a_size, b_size):
                # None is compatible with any size
                continue

            self.assertEqual(a_size, b_size,
                             '{} vs {}: column "{}" has size "{}" != "{}"'.format(
                                 relation_a, relation_b, a_name, a_size, b_size
                             ))

    def assertEquals(self, *args, **kwargs):
        # assertEquals is deprecated. This makes the warnings less chatty
        self.assertEqual(*args, **kwargs)

    def assertBetween(self, timestr, start, end=None):
        datefmt = '%Y-%m-%dT%H:%M:%S.%fZ'
        if end is None:
            end = datetime.utcnow()

        parsed = datetime.strptime(timestr, datefmt)

        self.assertLessEqual(start, parsed,
                             'parsed date {} happened before {}'.format(
                                 parsed,
                                 start.strftime(datefmt))
                             )
        self.assertGreaterEqual(end, parsed,
                                'parsed date {} happened after {}'.format(
                                    parsed,
                                    end.strftime(datefmt))
                                )


def use_profile(profile_name):
    """A decorator to declare a test method as using a particular profile.
    Handles both setting the nose attr and calling self.use_profile.

    Use like this:

    class TestSomething(DBIntegrationTest):
        @use_profile('postgres')
        def test_postgres_thing(self):
            self.assertEqual(self.adapter_type, 'postgres')

        @use_profile('snowflake')
        def test_snowflake_thing(self):
            self.assertEqual(self.adapter_type, 'snowflake')
    """

    def outer(wrapped):
        @getattr(pytest.mark, 'profile_' + profile_name)
        @wraps(wrapped)
        def func(self, *args, **kwargs):
            return wrapped(self, *args, **kwargs)

        # sanity check at import time
        assert _profile_from_test_name(wrapped.__name__) == profile_name
        return func

    return outer


class AnyFloat:
    """Any float. Use this in assertEqual() calls to assert that it is a float.
    """

    def __eq__(self, other):
        return isinstance(other, float)


class AnyString:
    """Any string. Use this in assertEqual() calls to assert that it is a string.
    """

    def __eq__(self, other):
        return isinstance(other, str)


class AnyStringWith:
    def __init__(self, contains=None):
        self.contains = contains

    def __eq__(self, other):
        if not isinstance(other, str):
            return False

        if self.contains is None:
            return True

        return self.contains in other

    def __repr__(self):
        return 'AnyStringWith<{!r}>'.format(self.contains)



from dbt.adapters.base import AdapterPlugin

from dbt.adapters.oracle.connections import OracleConnectionManager
from dbt.adapters.oracle.connections import OracleCredentials
from dbt.adapters.oracle.impl import OracleAdapter
from dbt.include import oracle

Plugin = AdapterPlugin(
    adapter=OracleAdapter,
    credentials=OracleCredentials,
    include_path=oracle.PACKAGE_PATH)



import decimal
import random
import unittest
from unittest.mock import MagicMock, call

import cx_Oracle
import dbt.flags as flags
from cx_Oracle import DatabaseError
from dbt.adapters.base.query_headers import MacroQueryStringSetter
from dbt.clients import agate_helper
from dbt.exceptions import ValidationException, DbtConfigError
from dbt.logger import GLOBAL_LOGGER as logger  # noqa
from dbt.parser import ParseResult
from dbt.task.debug import DebugTask

from dbt.adapters.oracle import OracleAdapter, Plugin as OraclePlugin
from .utils import *


class TestOracleAdapter(unittest.TestCase):

    def setUp(self):
        flags.STRICT_MODE = True
        project_cfg = {
            'name': 'X',
            'version': '0.1',
            'profile': 'test',
            'project-root': '/tmp/dbt/does-not-exist',
            'config-version': 2,
        }
        profile_cfg = {
            'outputs': {
                'test': {
                    'type': 'oracle',
                    'service': 'test',
                    'username': 'root',
                    'host': 'thishostshouldnotexist',
                    'password': 'password',
                    'port': 1521,
                    'schema': 'public'
                }
            },
            'target': 'test'
        }

        self.config = config_from_parts_or_dicts(project_cfg, profile_cfg)
        self._adapter = None

    @property
    def adapter(self):
        if self._adapter is None:
            self._adapter = OracleAdapter(self.config)
            inject_adapter(self._adapter, OraclePlugin)
        return self._adapter

    @mock.patch('dbt.adapters.oracle.connections.oracle')
    def test_acquire_connection_validations(self, oracle):
        try:
            connection = self.adapter.acquire_connection('dummy')
        except ValidationException as e:
            self.fail('got ValidationException: {}'.format(str(e)))
        except BaseException as e:
            self.fail('acquiring connection failed with unknown exception: {}'
                      .format(str(e)))
        self.assertEqual(connection.type, 'oracle')

        oracle.connect.assert_not_called()
        connection.handle
        oracle.connect.assert_called_once()

    @mock.patch('dbt.adapters.oracle.connections.oracle')
    def test_acquire_connection(self, oracle):
        connection = self.adapter.acquire_connection('dummy')

        oracle.connect.assert_not_called()
        connection.handle
        self.assertEqual(connection.state, 'open')
        self.assertNotEqual(connection.handle, None)
        oracle.connect.assert_called_once()

    def test_cancel_open_connections_empty(self):
        self.assertEqual(len(list(self.adapter.cancel_open_connections())), 0)

    def test_cancel_open_connections_master(self):
        key = self.adapter.connections.get_thread_identifier()
        self.adapter.connections.thread_connections[key] = mock_connection('master')
        self.assertEqual(len(list(self.adapter.cancel_open_connections())), 0)

    def test_cancel_open_connections_single(self):
        master = mock_connection('master')
        model = mock_connection('model')
        key = self.adapter.connections.get_thread_identifier()
        model.handle.get_backend_pid.return_value = 42
        self.adapter.connections.thread_connections.update({
            key: master,
            1: model,
        })
        self.assertEqual(len(list(self.adapter.cancel_open_connections())), 1)
        model.handle.cancel.assert_called_once()
        model.handle.add_query.assert_not_called()

    @mock.patch('dbt.adapters.oracle.connections.oracle')
    def test_changed_as_sysdba(self, oracle):
        self.config.credentials = self.config.credentials.replace(as_sysdba=True)
        connection = self.adapter.acquire_connection('dummy')

        oracle.connect.assert_not_called()
        connection.handle
        oracle.connect.assert_called_once_with(
            'root',
            'password',
            'thishostshouldnotexist:1521/test',
            mode=oracle.SYSDBA
        )

    @mock.patch('dbt.adapters.oracle.connections.oracle')
    def test_default_auth(self, oracle):
        connection = self.adapter.acquire_connection('dummy')

        oracle.connect.assert_not_called()
        connection.handle
        oracle.connect.assert_called_once_with(
            'root',
            'password',
            'thishostshouldnotexist:1521/test',
            mode=oracle.DEFAULT_AUTH
        )

    @mock.patch.object(OracleAdapter, 'execute_macro')
    @mock.patch.object(OracleAdapter, '_get_catalog_schemas')
    def test_get_catalog_various_schemas(self, mock_get_schemas, mock_execute):
        column_names = ['table_database', 'table_schema', 'table_name']
        rows = [
            ('dbt', 'foo', 'bar'),
            ('dbt', 'FOO', 'baz'),
            ('dbt', None, 'bar'),
            ('dbt', 'quux', 'bar'),
            ('dbt', 'skip', 'bar'),
        ]
        mock_execute.return_value = agate.Table(rows=rows,
                                                column_names=column_names)

        mock_get_schemas.return_value.items.return_value = [(mock.MagicMock(database='dbt'), {'foo', 'FOO', 'quux'})]

        mock_manifest = mock.MagicMock()
        mock_manifest.get_used_schemas.return_value = {('dbt', 'foo'),
                                                       ('dbt', 'quux')}

        catalog, exceptions = self.adapter.get_catalog(mock_manifest)
        self.assertEqual(
            set(map(tuple, catalog)),
            {('dbt', 'foo', 'bar'), ('dbt', 'FOO', 'baz'), ('dbt', 'quux', 'bar')}
        )
        self.assertEqual(exceptions, [])

    @mock.patch('dbt.adapters.oracle.connections.oracle')
    def test_split_on_semicolon(self, oracle: MagicMock):
        self.adapter.acquire_connection('test_semicolon')

        query1 = "SELECT * FROM DUAL"
        query2 = "SELECT 'TEST' FROM DUAL"
        multi_query = f'{query1};\n{query2}'

        oracle.connect.assert_not_called()
        self.adapter.execute(multi_query)
        oracle.assert_has_calls(
            [
                call.connect().cursor().execute(query1),
                call.connect().cursor().execute(query2)
            ],
            any_order=True
        )

    @mock.patch('dbt.adapters.oracle.connections.oracle')
    def test_dont_split_semicolon_comment(self, oracle: MagicMock):
        self.adapter.acquire_connection('test_semicolon')

        query1 = "SELECT * FROM DUAL"
        query2 = "SELECT 'TEST' FROM DUAL"
        multi_query = f'--{query1};\n{query2}'

        oracle.connect.assert_not_called()
        self.adapter.execute(multi_query)
        oracle.assert_has_calls(
            [
                call.connect().cursor().execute(multi_query)
            ],
            any_order=True
        )

    @mock.patch('dbt.adapters.oracle.connections.oracle')
    def test_complex_semicolon_split(self, oracle: MagicMock):
        self.adapter.acquire_connection('test_semicolon')

        query1 = "--SELECT * FROM DUAL;SELECT 'A' FROM DUAL WHERE 1 = 0;;;"
        query2 = "SELECT 'TEST' FROM DUAL"
        query3 = "SELECT 'A' FROM DUAL"
        query4 = "UNION ALL SELECT 'B' FROM DUAL"

        first_query = f'{query1};\n{query2}'
        second_query = f'{query3} {query4}'
        multi_query = f'{first_query};\n{second_query}'

        oracle.connect.assert_not_called()
        self.adapter.execute(multi_query)
        oracle.assert_has_calls(
            [
                call.connect().cursor().execute(first_query),
                call.connect().cursor().execute(second_query)
            ],
            any_order=True
        )


class TestConnectingOracleAdapter(unittest.TestCase):
    def setUp(self):
        self.target_dict = {
            'type': 'oracle',
            'service': 'test',
            'username': 'root',
            'host': 'thishostshouldnotexist',
            'password': 'password',
            'port': 1521,
            'schema': 'public'
        }

        profile_cfg = {
            'outputs': {
                'test': self.target_dict,
            },
            'target': 'test'
        }
        project_cfg = {
            'name': 'X',
            'version': '0.1',
            'profile': 'test',
            'project-root': '/tmp/dbt/does-not-exist',
            'quoting': {
                'identifier': False,
                'schema': True,
            },
            'config-version': 2,
        }

        self.config = config_from_parts_or_dicts(project_cfg, profile_cfg)

        self.handle = mock.MagicMock(spec=cx_Oracle.Connection)
        self.cursor = self.handle.cursor.return_value
        self.mock_execute = self.cursor.execute
        self.patcher = mock.patch('dbt.adapters.oracle.connections.oracle')
        self.oracle = self.patcher.start()

        self.load_patch = mock.patch('dbt.parser.manifest.make_parse_result')
        self.mock_parse_result = self.load_patch.start()
        self.mock_parse_result.return_value = ParseResult.rpc()

        self.oracle.connect.return_value = self.handle
        self.adapter = OracleAdapter(self.config)
        self.adapter._macro_manifest_lazy = load_internal_manifest_macros(self.config)
        self.adapter.connections.query_header = MacroQueryStringSetter(self.config, self.adapter._macro_manifest_lazy)

        self.qh_patch = mock.patch.object(self.adapter.connections.query_header, 'add')
        self.mock_query_header_add = self.qh_patch.start()
        self.mock_query_header_add.side_effect = lambda q: '/* dbt */\n{}'.format(q)
        self.adapter.acquire_connection()
        inject_adapter(self.adapter, OraclePlugin)

    def tearDown(self):
        # we want a unique self.handle every time.
        self.adapter.cleanup_connections()
        self.qh_patch.stop()
        self.patcher.stop()
        self.load_patch.stop()
        clear_plugin(OraclePlugin)

    # def test_quoting_on_drop_schema(self):
    #     relation = self.adapter.Relation.create(
    #         database='oracle', schema='test_schema',
    #         quote_policy=self.adapter.config.quoting,
    #     )
    #     self.adapter.drop_schema(relation)
    #
    #     self.mock_execute.assert_has_calls([
    #         mock.call('/* dbt */\ndrop user test_schema cascade', None)
    #     ])

    # def test_quoting_on_drop(self):
    #     relation = self.adapter.Relation.create(
    #         database='oracle',
    #         schema='test_schema',
    #         identifier='test_table',
    #         type='table',
    #         quote_policy=self.adapter.config.quoting,
    #     )
    #     self.adapter.drop_relation(relation)
    #     self.mock_execute.assert_any_call([
    #         mock.call('/* dbt */\ndrop table test_table', None)
    #     ])

    def test_quoting_on_truncate(self):
        relation = self.adapter.Relation.create(
            database='oracle',
            schema='test_schema',
            identifier='test_table',
            type='table',
            quote_policy=self.adapter.config.quoting,
        )
        self.adapter.truncate_relation(relation)
        self.mock_execute.assert_called_once_with('/* dbt */\ntruncate table test_schema.test_table')

    # def test_quoting_on_rename(self):
    #     random.seed(1)  # set the seed to obtain the same random number every time
    #     from_relation = self.adapter.Relation.create(
    #         database='oracle',
    #         schema='test_schema',
    #         identifier='table_a',
    #         type='table',
    #         quote_policy=self.adapter.config.quoting,
    #     )
    #     to_relation = self.adapter.Relation.create(
    #         database='oracle',
    #         schema='test_schema',
    #         identifier='table_b',
    #         type='table',
    #         quote_policy=self.adapter.config.quoting,
    #     )
    #
    #     self.adapter.rename_relation(
    #         from_relation=from_relation,
    #         to_relation=to_relation
    #     )
    #     self.mock_execute.assert_called_with(
    #         '/* dbt */\nBEGIN\n            test_schema.DBT_RENAME__17611(\'table_a\', \'table_b\');\n        END;'
    #     )

    def test_debug_connection_ok(self):
        DebugTask.validate_connection(self.target_dict)
        self.mock_execute.assert_has_calls([
            mock.call('/* dbt */\nselect 1 as id from dual')
        ])

    def test_debug_connection_fail_nopass(self):
        del self.target_dict['password']
        with self.assertRaises(DbtConfigError):
            DebugTask.validate_connection(self.target_dict)

    def test_connection_fail_select(self):
        self.mock_execute.side_effect = DatabaseError()
        with self.assertRaises(DbtConfigError):
            DebugTask.validate_connection(self.target_dict)
        self.mock_execute.assert_has_calls([
            mock.call('/* dbt */\nselect 1 as id from dual')
        ])


class TestOracleFilterCatalog(unittest.TestCase):
    def test__catalog_filter_table(self):
        manifest = mock.MagicMock()
        manifest.get_used_schemas.return_value = [['a', 'B'], ['a', '1234']]
        column_names = ['table_name', 'table_database', 'table_schema', 'something']
        rows = [
            ['foo', 'a', 'b', '1234'],  # include
            ['foo', 'a', '1234', '1234'],  # include, w/ table schema as str
            ['foo', 'c', 'B', '1234'],  # skip
            ['1234', 'A', 'B', '1234'],  # include, w/ table name as str
        ]
        table = agate.Table(
            rows, column_names, agate_helper.DEFAULT_TYPE_TESTER
        )

        result = OracleAdapter._catalog_filter_table(table, manifest)
        assert len(result) == 3
        for row in result.rows:
            assert isinstance(row['table_schema'], str)
            assert isinstance(row['table_database'], str)
            assert isinstance(row['table_name'], str)
            assert isinstance(row['something'], decimal.Decimal)


class TestPostgresAdapterConversions(TestAdapterConversions):
    def test_convert_text_type(self):
        dbt_additional_space = OracleAdapter.oracle_additional_text_space_for_dbt()
        rows = [
            ['', 'a1', 'stringval1'],
            ['', 'a2', 'stringvalasdfasdfasdfa'],
            ['', 'a3', 'stringval3'],
        ]
        agate_table = self._make_table_of(rows, agate.Text)
        oracle_text_type = 'VARCHAR2({} CHAR)'
        expected = [
            oracle_text_type.format(0 + dbt_additional_space),
            oracle_text_type.format(2 + dbt_additional_space),
            oracle_text_type.format(22 + dbt_additional_space)
        ]
        for col_idx, expect in enumerate(expected):
            assert OracleAdapter.convert_text_type(agate_table, col_idx) == expect

    def test_convert_number_type(self):
        rows = [
            ['', '23.98', '-1'],
            ['', '12.78', '-2'],
            ['', '79.41', '-3'],
        ]
        agate_table = self._make_table_of(rows, agate.Number)
        expected = ['NUMBER', 'NUMBER', 'NUMBER']
        for col_idx, expect in enumerate(expected):
            assert OracleAdapter.convert_number_type(agate_table, col_idx) == expect

    def test_convert_boolean_type(self):
        rows = [
            ['', 'false', 'true'],
            ['', 'false', 'false'],
            ['', 'false', 'true'],
        ]
        agate_table = self._make_table_of(rows, agate.Boolean)
        expected = ['NUMBER', 'NUMBER', 'NUMBER']
        for col_idx, expect in enumerate(expected):
            assert OracleAdapter.convert_boolean_type(agate_table, col_idx) == expect

    def test_convert_datetime_type(self):
        rows = [
            ['', '20190101T01:01:01Z', '2019-01-01 01:01:01'],
            ['', '20190102T01:01:01Z', '2019-01-01 01:01:01'],
            ['', '20190103T01:01:01Z', '2019-01-01 01:01:01'],
        ]
        agate_table = self._make_table_of(rows, [agate.DateTime, agate_helper.ISODateTime, agate.DateTime])
        expected = ['DATE', 'DATE', 'DATE']
        for col_idx, expect in enumerate(expected):
            assert OracleAdapter.convert_datetime_type(agate_table, col_idx) == expect

    def test_convert_date_type(self):
        rows = [
            ['', '2019-01-01', '2019-01-04'],
            ['', '2019-01-02', '2019-01-04'],
            ['', '2019-01-03', '2019-01-04'],
        ]
        agate_table = self._make_table_of(rows, agate.Date)
        expected = ['TIMESTAMP', 'TIMESTAMP', 'TIMESTAMP']
        for col_idx, expect in enumerate(expected):
            assert OracleAdapter.convert_date_type(agate_table, col_idx) == expect

    def test_convert_time_type(self):
        # dbt's default type testers actually don't have a TimeDelta at all.
        agate.TimeDelta
        rows = [
            ['', '120s', '10s'],
            ['', '3m', '11s'],
            ['', '1h', '12s'],
        ]
        agate_table = self._make_table_of(rows, agate.TimeDelta)
        expected = ['TIMESTAMP', 'TIMESTAMP', 'TIMESTAMP']
        for col_idx, expect in enumerate(expected):
            assert OracleAdapter.convert_time_type(agate_table, col_idx) == expect



import re
from typing import Optional, List, Tuple, Union

import agate
import sqlparse
from dbt.adapters.base import BaseRelation
from dbt.adapters.sql import SQLAdapter
from dbt.contracts.connection import AdapterResponse

from .connections import OracleConnectionManager
from .relation import OracleRelation, OracleColumn


class OracleAdapter(SQLAdapter):
    ConnectionManager = OracleConnectionManager
    Relation = OracleRelation
    Column = OracleColumn

    def date_function(self) -> str:
        return "SYSDATE"

    def debug_query(self) -> None:
        self.execute('select 1 as id from dual')

    def timestamp_add_sql(self, add_to: str, number: int = 1, interval: str = 'hour') -> str:
        # in oracle is "interval '1' hour" and not "interval '1 hour'"
        return f"{add_to} + interval '{number}' {interval}"

    @classmethod
    def convert_text_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        max_str_length = agate_table.aggregate(agate.MaxLength(col_idx)) + cls.oracle_additional_text_space_for_dbt()
        return f"VARCHAR2({max_str_length} CHAR)"

    @classmethod
    def oracle_additional_text_space_for_dbt(cls) -> int:
        return 20

    @classmethod
    def convert_number_type(
            cls, agate_table: agate.Table, col_idx: int
    ) -> str:
        return "NUMBER"

    @classmethod
    def convert_boolean_type(
            cls, agate_table: agate.Table, col_idx: int
    ) -> str:
        return "NUMBER"

    @classmethod
    def convert_datetime_type(
            cls, agate_table: agate.Table, col_idx: int
    ) -> str:
        return "DATE"

    @classmethod
    def convert_date_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "TIMESTAMP"

    @classmethod
    def convert_time_type(cls, agate_table: agate.Table, col_idx: int) -> str:
        return "TIMESTAMP"

    def quote(self, identifier):
        return identifier  # never quote on Oracle adapter

    # Oracle doesn't support 'AS' for table alias
    # Oracle EXCEPT operator is MINUS
    def get_rows_different_sql(self, relation_a: BaseRelation, relation_b: BaseRelation,
                               column_names: Optional[List[str]] = None,
                               except_operator: str = 'MINUS') -> str:
        """Generate SQL for a query that returns a single row with a two
                columns: the number of rows that are different between the two
                relations and the number of mismatched rows.
                """
        # This method only really exists for test reasons.
        names: List[str]
        if column_names is None:
            columns = self.get_columns_in_relation(relation_a)
            names = sorted((self.quote(c.name) for c in columns))
        else:
            names = sorted((self.quote(n) for n in column_names))
        columns_csv = ', '.join(names)

        sql = COLUMNS_EQUAL_SQL.format(
            columns=columns_csv,
            relation_a=str(relation_a),
            relation_b=str(relation_b),
            except_op=except_operator,
        )

        return sql

    def execute(self, sql: str, auto_begin: bool = False, fetch: bool = False) -> Tuple[
        Union[str, AdapterResponse], agate.Table]:

        last_call = None
        for query in OracleAdapter._oracle_query_parse(sql):
            last_call = super().execute(query, auto_begin, fetch)

        # force commit at the end, avoid bugs that are hard as fuck to catch (like insert in hooks)
        # may be a problem?
        self.connections.add_commit_query()

        return last_call

    @staticmethod
    def _oracle_query_parse(sql):
        sql_list = []

        # plsql check
        # for plsql cx_Oracle needs the ending ;
        if 'end;' in sql.lower():
            return [sql]

        # split multiple statements, cx_Oracle can execute one statement at time

        for sql_query in sqlparse.split(sql):

            if sql_query.endswith(';'):
                sql_query = sql_query[:-1]

            if sql_query is not None and sql_query.strip() != '':
                sql_list.append(sql_query)

        return sql_list


# Oracle doesn't support 'AS' for table alias
COLUMNS_EQUAL_SQL = '''
with diff_count as (
    SELECT
        1 as id,
        COUNT(*) as num_missing 
    FROM (
            (SELECT {columns} FROM {relation_a} {except_op}
             SELECT {columns} FROM {relation_b})
             UNION ALL
            (SELECT {columns} FROM {relation_b} {except_op}
             SELECT {columns} FROM {relation_a})
        ) a
), oracle_table_a as (
    SELECT COUNT(*) as num_rows FROM {relation_a}
), oracle_table_b as (
    SELECT COUNT(*) as num_rows FROM {relation_b}
), row_count_diff as (
    select
        1 as id,
        oracle_table_a.num_rows - oracle_table_b.num_rows as difference
    from oracle_table_a, oracle_table_b
)
select
    row_count_diff.difference as row_count_difference,
    diff_count.num_missing as num_mismatched
from row_count_diff
join diff_count using (id)
'''.strip()



from dataclasses import dataclass
from typing import Optional, Iterator, Tuple, ClassVar, Dict, Any

from dbt.adapters.base import BaseRelation, Column
from dbt.contracts.relation import ComponentName, Policy
from dbt.exceptions import RuntimeException


@dataclass(frozen=True, eq=False, repr=False)
class OracleRelation(BaseRelation):
    def __post_init__(self):
        if (self.identifier is not None and self.type is not None and
                len(self.identifier) > self.relation_max_name_length()):
            raise RuntimeException(
                f"Relation name '{self.identifier}' "
                f"is longer than {self.relation_max_name_length()} characters"
            )

    @classmethod
    def get_default_include_policy(cls) -> Policy:
        return Policy(database=False)  # oracle use schema.object_name, not database.schema.object_name

    @classmethod
    def get_default_quote_policy(cls) -> Policy:
        return Policy(database=False, schema=False, identifier=False)  # usually identifiers in oracle are not quotes

    def _render_iterator(self) -> Iterator[Tuple[Optional[ComponentName], Optional[str]]]:
        self._force_oracle_policies()
        return super()._render_iterator()

    def _is_exactish_match(self, field: ComponentName, value: str) -> bool:
        self._force_oracle_policies()
        return self.path.get_lowered_part(field) == value.lower()  # always return check on lowered parts (ignore case)

    # Oracle doesn't support table_name that starts with _
    @staticmethod
    def add_ephemeral_prefix(name: str):
        return f'o{BaseRelation.add_ephemeral_prefix(name)}'

    @staticmethod
    def relation_max_name_length():
        # oracle max is 128 from DB 12.2, 30 is some spacing for DBT suffixes
        return 128 - 30

    def _force_oracle_policies(self):
        object.__setattr__(self, 'include_policy', self.get_default_include_policy())  # force oracle include policy
        object.__setattr__(self, 'quote_policy', self.get_default_quote_policy())  # force oracle quote policy


class OracleColumn(Column):
    TYPE_LABELS: ClassVar[Dict[str, str]] = {
        # in base class
        'STRING': 'VARCHAR2',
        'TIMESTAMP': 'TIMESTAMP',
        'FLOAT': 'NUMBER',
        'INTEGER': 'NUMBER',

        # all text types conversions
        'TEXT': 'VARCHAR2',
        'CHARACTER VARYING': 'VARCHAR2',
        'CHARACTER': 'VARCHAR2',
        'VARCHAR': 'VARCHAR2',

        # all integer/float types conversions
        'REAL': 'NUMBER',
        'FLOAT4': 'NUMBER',
        'DOUBLE PRECISION': 'NUMBER',
        'FLOAT8': 'NUMBER',
        'SMALLINT': 'NUMBER',
        'BIGINT': 'NUMBER',
        'SMALLSERIAL': 'NUMBER',
        'SERIAL': 'NUMBER',
        'BIGSERIAL': 'NUMBER',
        'INT2': 'NUMBER',
        'INT4': 'NUMBER',
        'INT8': 'NUMBER',
        'SERIAL2': 'NUMBER',
        'SERIAL4': 'NUMBER',
        'SERIAL8': 'NUMBER',
        'NUMERIC': 'NUMBER',
        'DECIMAL': 'NUMBER'
    }

    @property
    def data_type(self) -> str:
        if self.is_string():
            return OracleColumn.string_type(self.string_size())
        elif self.is_numeric():
            return OracleColumn.numeric_type(self.dtype, self.numeric_precision,
                                             self.numeric_scale)
        else:
            return self.dtype

    def is_string(self) -> bool:
        return self.dtype.lower() in ['varchar2', 'char']

    def is_float(self):
        return self.dtype.lower() == 'number'

    def is_integer(self) -> bool:
        return self.dtype.lower() == 'number'

    def is_numeric(self) -> bool:
        return self.dtype.lower() == 'number'

    def literal(self, value: Any) -> str:
        return "CAST({} AS {})".format(value, self.data_type)

    @classmethod
    def string_type(cls, size: int) -> str:
        return "VARCHAR2({} CHAR)".format(size)



"""
COPIED FROM DBT REPO (dbt repo/test/unit/utils.py)

Unit test utility functions.
Note that all imports should be inside the functions to avoid import/mocking
issues.
"""
import os
import string
from unittest import TestCase
from unittest import mock

import agate
import pytest
from hologram import ValidationError


def normalize(path):
    """On windows, neither is enough on its own:
    >>> normcase('C:\\documents/ALL CAPS/subdir\\..')
    'c:\\documents\\all caps\\subdir\\..'
    >>> normpath('C:\\documents/ALL CAPS/subdir\\..')
    'C:\\documents\\ALL CAPS'
    >>> normpath(normcase('C:\\documents/ALL CAPS/subdir\\..'))
    'c:\\documents\\all caps'
    """
    return os.path.normcase(os.path.normpath(path))


class Obj:
    which = 'blah'
    single_threaded = False


def mock_connection(name, state='open'):
    conn = mock.MagicMock()
    conn.name = name
    conn.state = state
    return conn


def profile_from_dict(profile, profile_name, cli_vars='{}'):
    from dbt.config import Profile
    from dbt.config.renderer import ProfileRenderer
    from dbt.context.base import generate_base_context
    from dbt.config.utils import parse_cli_vars
    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)

    renderer = ProfileRenderer(generate_base_context(cli_vars))
    return Profile.from_raw_profile_info(
        profile,
        profile_name,
        renderer,
    )


def project_from_dict(project, profile, packages=None, selectors=None, cli_vars='{}'):
    from dbt.context.target import generate_target_context
    from dbt.config import Project
    from dbt.config.renderer import DbtProjectYamlRenderer
    from dbt.config.utils import parse_cli_vars
    if not isinstance(cli_vars, dict):
        cli_vars = parse_cli_vars(cli_vars)

    renderer = DbtProjectYamlRenderer(generate_target_context(profile, cli_vars))

    project_root = project.pop('project-root', os.getcwd())

    return Project.render_from_dict(
        project_root, project, packages, selectors, renderer
    )


def config_from_parts_or_dicts(project, profile, packages=None, selectors=None, cli_vars='{}'):
    from dbt.config import Project, Profile, RuntimeConfig
    from copy import deepcopy

    if isinstance(project, Project):
        profile_name = project.profile_name
    else:
        profile_name = project.get('profile')

    if not isinstance(profile, Profile):
        profile = profile_from_dict(
            deepcopy(profile),
            profile_name,
            cli_vars,
        )

    if not isinstance(project, Project):
        project = project_from_dict(
            deepcopy(project),
            profile,
            packages,
            selectors,
            cli_vars,
        )

    args = Obj()
    args.vars = cli_vars
    args.profile_dir = '/dev/null'
    return RuntimeConfig.from_parts(
        project=project,
        profile=profile,
        args=args
    )


def inject_plugin(plugin):
    from dbt.adapters.factory import FACTORY
    key = plugin.adapter.type()
    FACTORY.plugins[key] = plugin


def inject_plugin_for(config):
    # from dbt.adapters.postgres import Plugin, PostgresAdapter
    from dbt.adapters.factory import FACTORY
    FACTORY.load_plugin(config.credentials.type)
    adapter = FACTORY.get_adapter(config)
    return adapter


def inject_adapter(value, plugin):
    """Inject the given adapter into the adapter factory, so your hand-crafted
    artisanal adapter will be available from get_adapter() as if dbt loaded it.
    """
    inject_plugin(plugin)
    from dbt.adapters.factory import FACTORY
    key = value.type()
    FACTORY.adapters[key] = value


def clear_plugin(plugin):
    from dbt.adapters.factory import FACTORY
    key = plugin.adapter.type()
    FACTORY.plugins.pop(key, None)
    FACTORY.adapters.pop(key, None)


class ContractTestCase(TestCase):
    ContractType = None

    def setUp(self):
        self.maxDiff = None
        super().setUp()

    def assert_to_dict(self, obj, dct):
        self.assertEqual(obj.to_dict(), dct)

    def assert_from_dict(self, obj, dct, cls=None):
        if cls is None:
            cls = self.ContractType
        self.assertEqual(cls.from_dict(dct), obj)

    def assert_symmetric(self, obj, dct, cls=None):
        self.assert_to_dict(obj, dct)
        self.assert_from_dict(obj, dct, cls)

    def assert_fails_validation(self, dct, cls=None):
        if cls is None:
            cls = self.ContractType

        with self.assertRaises(ValidationError):
            cls.from_dict(dct)


def assert_to_dict(obj, dct):
    assert obj.to_dict() == dct


def assert_from_dict(obj, dct, cls=None):
    if cls is None:
        cls = obj.__class__
    assert cls.from_dict(dct) == obj


def assert_symmetric(obj, dct, cls=None):
    assert_to_dict(obj, dct)
    assert_from_dict(obj, dct, cls)


def assert_fails_validation(dct, cls):
    with pytest.raises(ValidationError):
        cls.from_dict(dct)


def generate_name_macros(package):
    from dbt.contracts.graph.parsed import ParsedMacro
    from dbt.node_types import NodeType
    name_sql = {}
    for component in ('database', 'schema', 'alias'):
        if component == 'alias':
            source = 'node.name'
        else:
            source = f'target.{component}'
        name = f'generate_{component}_name'
        sql = f'{{% macro {name}(value, node) %}} {{% if value %}} {{{{ value }}}} {{% else %}} {{{{ {source} }}}} {{% endif %}} {{% endmacro %}}'
        name_sql[name] = sql

    for name, sql in name_sql.items():
        pm = ParsedMacro(
            name=name,
            resource_type=NodeType.Macro,
            unique_id=f'macro.{package}.{name}',
            package_name=package,
            original_file_path=normalize('macros/macro.sql'),
            root_path='./dbt_modules/root',
            path=normalize('macros/macro.sql'),
            macro_sql=sql,
        )
        yield pm


class TestAdapterConversions(TestCase):
    def _get_tester_for(self, column_type):
        from dbt.clients import agate_helper
        if column_type is agate.TimeDelta:  # dbt never makes this!
            return agate.TimeDelta()

        for instance in agate_helper.DEFAULT_TYPE_TESTER._possible_types:
            if type(instance) is column_type:
                return instance

        raise ValueError(f'no tester for {column_type}')

    def _make_table_of(self, rows, column_types):
        column_names = list(string.ascii_letters[:len(rows[0])])
        if isinstance(column_types, type):
            column_types = [self._get_tester_for(column_types) for _ in column_names]
        else:
            column_types = [self._get_tester_for(typ) for typ in column_types]
        table = agate.Table(rows, column_names=column_names, column_types=column_types)
        return table


def MockMacro(package, name='my_macro', **kwargs):
    from dbt.contracts.graph.parsed import ParsedMacro
    from dbt.node_types import NodeType

    mock_kwargs = dict(
        resource_type=NodeType.Macro,
        package_name=package,
        unique_id=f'macro.{package}.{name}',
        original_file_path='/dev/null',
    )

    mock_kwargs.update(kwargs)

    macro = mock.MagicMock(
        spec=ParsedMacro,
        **mock_kwargs
    )
    macro.name = name
    return macro


def MockMaterialization(package, name='my_materialization', adapter_type=None, **kwargs):
    if adapter_type is None:
        adapter_type = 'default'
    kwargs['adapter_type'] = adapter_type
    return MockMacro(package, f'materialization_{name}_{adapter_type}', **kwargs)


def MockGenerateMacro(package, component='some_component', **kwargs):
    name = f'generate_{component}_name'
    return MockMacro(package, name=name, **kwargs)


def MockSource(package, source_name, name, **kwargs):
    from dbt.node_types import NodeType
    from dbt.contracts.graph.parsed import ParsedSourceDefinition
    src = mock.MagicMock(
        __class__=ParsedSourceDefinition,
        resource_type=NodeType.Source,
        source_name=source_name,
        package_name=package,
        unique_id=f'source.{package}.{source_name}.{name}',
        search_name=f'{source_name}.{name}',
        **kwargs
    )
    src.name = name
    return src


def MockNode(package, name, resource_type=None, **kwargs):
    from dbt.node_types import NodeType
    from dbt.contracts.graph.parsed import ParsedModelNode, ParsedSeedNode
    if resource_type is None:
        resource_type = NodeType.Model
    if resource_type == NodeType.Model:
        cls = ParsedModelNode
    elif resource_type == NodeType.Seed:
        cls = ParsedSeedNode
    else:
        raise ValueError(f'I do not know how to handle {resource_type}')
    node = mock.MagicMock(
        __class__=cls,
        resource_type=resource_type,
        package_name=package,
        unique_id=f'{str(resource_type)}.{package}.{name}',
        search_name=name,
        **kwargs
    )
    node.name = name
    return node


def MockDocumentation(package, name, **kwargs):
    from dbt.node_types import NodeType
    from dbt.contracts.graph.parsed import ParsedDocumentation
    doc = mock.MagicMock(
        __class__=ParsedDocumentation,
        resource_type=NodeType.Documentation,
        package_name=package,
        search_name=name,
        unique_id=f'{package}.{name}',
        **kwargs
    )
    doc.name = name
    return doc


def load_internal_manifest_macros(config, macro_hook=lambda m: None):
    from dbt.adapters.factory import get_include_paths
    from dbt.parser.manifest import ManifestLoader
    paths = get_include_paths(config.credentials.type)
    projects = {k: v for k, v in config.load_dependencies().items() if k.startswith('dbt')}
    loader = ManifestLoader(config, projects, macro_hook)
    return loader.load_only_macros()


def dict_replace(dct, **kwargs):
    dct = dct.copy()
    dct.update(kwargs)
    return dct


def replace_config(n, **kwargs):
    return n.replace(
        config=n.config.replace(**kwargs),
        unrendered_config=dict_replace(n.unrendered_config, **kwargs),
    )



import time
from contextlib import contextmanager
from dataclasses import dataclass
from typing import ContextManager, Any, Optional, Tuple, Union

import cx_Oracle as oracle
from dbt.adapters.base import Credentials
from dbt.adapters.sql import SQLConnectionManager
from dbt.contracts.connection import Connection, AdapterResponse
from dbt.exceptions import RuntimeException, FailedToConnectException
from dbt.logger import GLOBAL_LOGGER as logger


@dataclass
class OracleCredentials(Credentials):
    username: str
    password: str
    host: str
    port: int
    as_sysdba: Optional[bool] = False
    nls_date_format: Optional[str] = None

    _ALIASES = {
        'service': 'database'
    }

    @property
    def type(self):
        return 'oracle'

    def _connection_keys(self):
        return 'host', 'port', 'database', 'schema', 'username', 'as_sysdba', 'nls_date_format'

    @property
    def connection_string(self):
        return f"{self.username}@{self.host}:{self.port}/{self.database}{' AS SYSDBA' if self.as_sysdba else ''}"


class OracleConnectionManager(SQLConnectionManager):
    TYPE = 'oracle'

    @classmethod
    def get_response(cls, cursor: Any) -> Union[AdapterResponse, str]:
        return AdapterResponse(_message='OK', code=cursor.statement, rows_affected=cursor.rowcount)

    def cancel(self, connection: Connection):
        if connection.handle is not None:
            connection.handle.cancel()

    @classmethod
    def get_status(cls, cursor: Any) -> str:
        return f'OK - {cursor.rowcount} rows affected'

    @contextmanager
    def exception_handler(self, sql: str) -> ContextManager:
        try:
            yield
        except Exception as e:
            logger.error("Error running SQL: {}".format(sql))
            logger.error("Rolling back transaction.")
            self.rollback_if_open()

            if isinstance(e, RuntimeException):
                # during a sql query, an internal to dbt exception was raised.
                # this sounds a lot like a signal handler and probably has
                # useful information, so raise it without modification.
                raise

            raise RuntimeException(e) from e

    @classmethod
    def open(cls, connection: Connection) -> Connection:
        if connection.state == 'open':
            logger.debug('Connection is already open, skipping open.')
            return connection

        credentials = connection.credentials

        try:
            connection_string = '{}:{}/{}'
            handle = oracle.connect(
                credentials.username,
                credentials.password,
                connection_string.format(credentials.host, credentials.port, credentials.database),
                mode=oracle.SYSDBA if credentials.as_sysdba else oracle.DEFAULT_AUTH
            )
            connection.state = 'open'
            connection.handle = handle

            if credentials.nls_date_format is not None:
                handle.cursor().execute(f"ALTER SESSION SET NLS_DATE_FORMAT = '{credentials.nls_date_format}'")
                handle.cursor().execute(f"ALTER SESSION SET NLS_TIMESTAMP_FORMAT = '{credentials.nls_date_format}XFF'")
                handle.cursor().execute(f"ALTER SESSION SET NLS_TIMESTAMP_TZ_FORMAT = '{credentials.nls_date_format}XFF TZR'")


        except Exception as e:
            logger.debug(
                f"Got an error when attempting to open a oracle connection. Connection string = {credentials.connection_string}. Error: {e}"
            )

            connection.handle = None
            connection.state = 'fail'

            raise FailedToConnectException(f"{str(e)} using connection string {credentials.connection_string}")

        return connection

    def add_query(self, sql: str, auto_begin: bool = True, bindings: Optional[Any] = None,
                  abridge_sql_log: bool = False) -> Tuple[Connection, Any]:
        connection = self.get_thread_connection()
        if auto_begin and connection.transaction_open is False:
            self.begin()

        logger.debug('Using {} connection "{}".'
                     .format(self.TYPE, connection.name))

        with self.exception_handler(sql):

            if abridge_sql_log:
                log_sql = '{}...'.format(sql[:512])
            else:
                log_sql = sql

            logger.debug(
                'On {connection_name}: {sql}',
                connection_name=connection.name,
                sql=log_sql,
            )
            pre = time.time()

            cursor = connection.handle.cursor()

            execute_many = False
            if 'insert' in sql.lower() and bindings is not None:
                execute_many = True

            # cx_Oracle doesn't like if bindings is None and it is passed to execute. good job oracle
            if bindings is not None:
                if execute_many:
                    cursor.executemany(sql, bindings)
                else:
                    cursor.execute(sql, bindings)
            else:
                cursor.execute(sql)

            logger.debug(
                "SQL status: {status} in {elapsed:0.2f} seconds",
                status=self.get_status(cursor),
                elapsed=(time.time() - pre)
            )
            return connection, cursor

    def add_begin_query(self):
        self.get_thread_connection().handle.begin()

    def add_commit_query(self):
        self.get_thread_connection().handle.commit()



#!/usr/bin/env python
from setuptools import find_packages
from setuptools import setup

package_name = "oracle4dbt"
package_version = "0.0.1"
description = """The oracle adapter plugin for dbt (data build tool)"""
author = "rsella"
author_email = "rsella94@gmail.com"
github = "https://github.com/rsella/oracle4dbt"

with open('README.md', 'r', encoding='UTF-8') as f:
    long_description = f.read()

setup(
    name=package_name,
    version=package_version,
    description=description,
    long_description=long_description,
    long_description_content_type="text/markdown",
    author=author,
    author_email=author_email,
    url=github,
    packages=find_packages('.', exclude=['test']),
    package_data={
        'dbt': [
            'include/oracle/macros/*.sql',
            'include/oracle/dbt_project.yml',
        ]
    },
    install_requires=[
        "dbt-core>=0.19.0",
        "cx_Oracle"
    ],
    python_requires='>=3.5'
)
